#  Copyright (c) 2021 zfit
import itertools

import numpy as np
import pytest
import tensorflow_probability as tfp
from matplotlib import pyplot as plt

import zfit
from zfit import z
from zfit.core.interfaces import ZfitParameter


@pytest.mark.skip()  # copy not yet implemented
def test_copy_kde():
    size = 500
    data = np.random.normal(size=size, loc=2, scale=3)

    limits = (-15, 5)
    obs = zfit.Space("obs1", limits=limits)
    kde_adaptive = zfit.pdf.GaussianKDE1DimV1(data=data, bandwidth='adaptive',
                                              obs=obs,
                                              truncate=False)
    kde_adaptive.copy()


def create_pdf_sample(npoints=500, upper=None):
    if upper is None:
        upper = 11
    limits = (-13, upper)
    obs = zfit.Space("obs1", limits=limits)
    cb = zfit.pdf.CrystalBall(mu=2, sigma=3, alpha=1, n=25, obs=obs)
    gauss = zfit.pdf.Gauss(mu=-5, sigma=2.5, obs=obs)
    pdf = zfit.pdf.SumPDF([cb, gauss], fracs=0.8)
    data = pdf.sample(n=npoints)
    return data, obs, pdf


def test_grid_kde():
    data, obs, pdf = create_pdf_sample()
    with pytest.raises(ValueError):
        zfit.pdf.KDE1DimGrid(obs=obs, data=data, bandwidth='eoaue')


def test_exact_kde():
    data, obs, pdf = create_pdf_sample()
    with pytest.raises(ValueError):
        zfit.pdf.KDE1DimExact(obs=obs, data=data, bandwidth='eoaue')


def create_kde(kdetype=None, npoints=1500, cfgonly=False, nonly=False, full=True, upper=None, legacy=True,
               padding=False):
    import tensorflow as tf

    import zfit
    import zfit.z.numpy as znp

    npoints_lim = min(npoints, 2500)

    class StudentT(tfp.distributions.StudentT):

        def __init__(self, *args, **kwargs):
            super().__init__(*args, df=5, **kwargs)

    configs = []
    with tf.init_scope():
        if cfgonly or nonly:
            hparam1 = 0.9
        else:
            hparam1 = zfit.Parameter("h1", 0.9)
    comb = itertools.product([
        ('bandwidth', h) for h in [None, hparam1, 'adaptive', 'silverman', 'scott']
    ],
        [
            ('truncate', bo) for bo in [False, True]
        ],
        [
            ('type', zfit.pdf.GaussianKDE1DimV1)
        ], [('npoints', npoints_lim)]
    )
    if full or legacy:
        if legacy:
            comb = [next(comb)]
        configs.extend(comb)

    with tf.init_scope():
        if cfgonly or nonly:
            hparam2 = 0.9
        else:
            hparam2 = zfit.Parameter("h2", 0.92)
    comb = itertools.product([
        ('bandwidth', h) for h in
        [None, hparam2, 'adaptive_zfit', 'adaptive_geom', 'adaptive_std', 'silverman', 'scott', 'isj']
    ],
        [
            ('weights', weight) for weight in [None, znp.ones(shape=npoints_lim)]
        ], [
            ('padding', padding)
        ],
        [
            ('kernel', dist) for dist in [None, StudentT]
        ], [
            ('type', zfit.pdf.KDE1DimExact)
        ], [('npoints', npoints_lim)]
    )
    if not full:
        comb = [next(comb)]
    configs.extend(comb)

    # Grid PDFs

    comb = itertools.product([
        ('bandwidth', h) for h in ['adaptive_geom', None, hparam2, 'silverman', 'scott', 'adaptive_zfit']
    ],
        [
            ('weights', weight) for weight in [None, znp.ones(shape=npoints)]
        ], [
            ('padding', padding)
        ],
        [
            ('kernel', dist) for dist in [None, StudentT]
        ], [
            ('type', zfit.pdf.KDE1DimGrid)
        ],
        [
            ('num_grid_points', n) for n in [None, 800]
        ],
        [
            ('binning_method', method) for method in ['linear', 'simple']
        ],

    )

    if not full:
        comb = [next(comb)]
    configs.extend(comb)

    # FFT combinations

    comb = itertools.product([
        ('bandwidth', h) for h in [None, 'scott']  # TODO: adaptive?
    ],
        [
            ('weights', weight) for weight in [None, znp.ones(shape=npoints)]
        ], [
            ('padding', padding)
        ],
        [
            ('kernel', dist) for dist in [None, StudentT]
        ], [
            ('type', zfit.pdf.KDE1DimFFT)
        ],
        [
            ('num_grid_points', n) for n in [None, 800]
        ],
        [
            ('binning_method', method) for method in ['linear', 'simple']
        ],

    )

    if not full:
        comb = [next(comb)]
    configs.extend(comb)

    # IFJ combinations

    comb = itertools.product(
        [
            ('weights', weight) for weight in [None, znp.ones(shape=npoints)]
        ], [
            ('padding', padding)
        ],
        [
            ('type', zfit.pdf.KDE1DimISJ)
        ],
        [
            ('num_grid_points', n) for n in [None, 800]
        ],
        [
            ('binning_method', method) for method in ['linear', 'simple']
        ],

    )

    if not full:
        comb = [next(comb)]
    configs.extend(comb)

    # end config builder
    if nonly:
        return len(configs)
    cfg = configs[kdetype]
    cfg = dict(cfg)
    if cfgonly:
        return cfg
    _ = cfg.pop('npoints', None)

    constructor = cfg.pop('type')
    if constructor in (zfit.pdf.GaussianKDE1DimV1, zfit.pdf.KDE1DimExact):
        npoints = npoints_lim

    data, obs, pdf = create_pdf_sample(npoints, upper=upper)

    cfg['data'] = data
    cfg['obs'] = obs

    kde = constructor(**cfg)

    return kde, pdf, data.value()[:, 0]


def _get_print_kde(**kwargs):
    kdes = create_kde(**kwargs)
    return kdes


@pytest.hookimpl
def pytest_generate_tests(metafunc):
    if metafunc.function.__name__ == 'test_all_kde':
        full = metafunc.config.getoption("--longtests")
        full = full or metafunc.config.getoption("--longtests-kde")

        metafunc.parametrize('kdetype', [i for i in range(_get_print_kde(nonly=True, full=full))])


# @pytest.mark.flaky(3)
@pytest.mark.parametrize('jit', [
    False,
    True
])
@pytest.mark.parametrize('npoints', [1100, 500_000])
def test_all_kde(kdetype, npoints, jit, request):
    import zfit
    full = request.config.getoption("--longtests")
    full = full or request.config.getoption("--longtests-kde")
    cfg = create_kde(kdetype=kdetype, npoints=npoints, cfgonly=True, full=full)
    print(cfg)
    kdetype = kdetype, npoints
    if jit:
        run_jit = z.function(_run)
        expected_integral, integral, name, prob, prob_true, rel_tol, sample, sample2, x, name, data = run_jit(kdetype,
                                                                                                              full=full)
    else:

        expected_integral, integral, name, prob, prob_true, rel_tol, sample, sample2, x, name, data = _run(kdetype,
                                                                                                           full=full)

    expected_integral = zfit.run(expected_integral)

    if not jit:
        import matplotlib.pyplot as plt
        plt.figure()
        kernel_used = cfg.get('kernel')
        if kernel_used is not None:
            kernel_print = f", kernel={kernel_used.__name__}"
        else:
            kernel_print = ''
        bandwidth_printready = cfg.get('bandwidth')
        if isinstance(bandwidth_printready, ZfitParameter):
            bandwidth_printready = f'Param({float(bandwidth_printready.value())}'
        if bandwidth_printready is None:
            bandwidth_print = ""
        else:
            bandwidth_print = f", h={bandwidth_printready}"

        num_grid_points = cfg.get('num_grid_points')
        if num_grid_points is not None:
            grid_print = f', grid={num_grid_points}'
        else:
            grid_print = ''
        binning_method = cfg.get('binning_method')
        if binning_method is not None:
            binning_print = f', binning={binning_method}'
        else:
            binning_print = ''
        weights = cfg.get('weights')
        if weights is not None:
            weights_print = ', weighted'
        else:
            weights_print = ''
        npoints_plot = cfg.get('npoints', kdetype[1])
        plt.title(
            f"{name} with {npoints_plot} points{weights_print}{bandwidth_print}{kernel_print}{grid_print}{binning_print}",
            fontsize=10)
        plt.plot(x, prob, label=f'KDE')
        plt.plot(x, prob_true, label='true PDF')
        data_np = zfit.run(data)
        plt.hist(data_np, bins=40, density=True, alpha=0.3, label="Kernel points")
        plt.legend()

    abs_tol = 0.005 if kdetype[1] > 3000 else 0.03
    tolfac = 6 if not cfg['type'] == tfp.distributions.Normal else 1
    if cfg.get('binning_method') == 'simple':
        tolfac *= 6
    assert zfit.run(integral) == pytest.approx(expected_integral, abs=abs_tol * tolfac)
    assert tuple(sample.shape) == (1, 1)
    assert tuple(sample2.shape) == (1500, 1)
    assert prob.shape.rank == 1
    assert prob.shape[0] == x.shape[0]
    assert np.mean(prob - prob_true) < 0.07 * tolfac
    # make sure that on average, most values are close
    assert np.mean((prob / prob_true)[prob_true > np.mean(prob_true)] ** 2) == pytest.approx(1, abs=0.1 * tolfac)
    rtol = 0.05
    np.testing.assert_allclose(prob, prob_true, rtol=rtol, atol=0.01 * tolfac)


def _run(kdetype, full, upper=None, legacy=True, padding=False):
    from zfit.z import numpy as znp
    kde, pdf, data = create_kde(*kdetype, full=full, upper=upper, legacy=legacy, padding=padding)
    integral = kde.integrate(limits=kde.space, norm_range=(-3, 2))
    expected_integral = kde.integrate(limits=kde.space, norm_range=(-3, 2))
    rel_tol = 0.04
    sample = kde.sample(1).value()
    sample2 = kde.sample(1500).value()
    x = znp.linspace(*kde.space.limit1d, 30000)
    prob = kde.pdf(x)

    prob_true = pdf.pdf(x)

    return expected_integral, integral, kde.name, prob, prob_true, rel_tol, sample, sample2, x, kde.name, data


@pytest.mark.parametrize('kdetype', [i for i in range(_get_print_kde(nonly=True, full=False, legacy=False))])
@pytest.mark.parametrize('npoints', [1100, 500_000])
@pytest.mark.parametrize('upper', [-4, -1, 3])
def test_kde_border(kdetype, npoints, upper):
    import zfit
    cfg = create_kde(kdetype=kdetype, npoints=npoints, cfgonly=True, full=False, legacy=False)
    kdetype = kdetype, npoints

    expected_integral, integral, name, prob, prob_true, rel_tol, sample, sample2, x, name, data = _run(kdetype,
                                                                                                       full=False,
                                                                                                       upper=upper,
                                                                                                       legacy=False,
                                                                                                       padding=True)

    plt.figure()
    kernel_used = cfg.get('kernel')
    if kernel_used is not None:
        kernel_print = f", kernel={kernel_used.__name__}"
    else:
        kernel_print = ''
    bandwidth_printready = cfg.get('bandwidth')
    if isinstance(bandwidth_printready, ZfitParameter):
        bandwidth_printready = f'Param({float(bandwidth_printready.value())}'
    if bandwidth_printready is None:
        bandwidth_print = ""
    else:
        bandwidth_print = f", h={bandwidth_printready}"

    num_grid_points = cfg.get('num_grid_points')
    if num_grid_points is not None:
        grid_print = f', grid={num_grid_points}'
    else:
        grid_print = ''
    binning_method = cfg.get('binning_method')
    if binning_method is not None:
        binning_print = f', binning={binning_method}'
    else:
        binning_print = ''
    weights = cfg.get('weights')
    if weights is not None:
        weights_print = ', weighted'
    else:
        weights_print = ''
    npoints_plot = cfg.get('npoints', kdetype[1])
    plt.title(
        f"{name}, {npoints_plot} points, upper={upper}{weights_print}{bandwidth_print}"
        f"{kernel_print}{grid_print}{binning_print}",
        fontsize=10)
    plt.plot(x, prob, label=f'KDE')
    plt.plot(x, prob_true, label='true PDF')
    data_np = zfit.run(data)
    plt.hist(data_np, bins=40, density=True, alpha=0.3, label="Kernel points")
    plt.legend()

    integral = zfit.run(integral)
    expected_integral = zfit.run(expected_integral)
    abs_tol = 0.05
    assert zfit.run(integral) == pytest.approx(expected_integral, abs=abs_tol)
    assert tuple(sample.shape) == (1, 1)
    assert tuple(sample2.shape) == (1500, 1)
    assert prob.shape.rank == 1
    assert np.mean(prob - prob_true) < 0.07
    # make sure that on average, most values are close
    assert np.mean((prob / prob_true)[prob_true > np.mean(prob_true)] ** 2) == pytest.approx(1, abs=0.15)
    rtol = 0.05
    np.testing.assert_allclose(prob, prob_true, rtol=rtol, atol=0.07)



import tensorflow as tf
from tf_quant_finance.math import root_search
from tf_quant_finance.math.root_search.utils import \
    default_relative_root_tolerance

from ..settings import ztypes


def find_practical_support_bandwidth(kernel, bandwidth, absolute_tolerance=10e-5):
    """Return the kernel support for practical purposes.

    Used to find a support value for computations for kernel functions without finite (bounded) support.
    """
    absolute_root_tolerance = 1e-3
    relative_root_tolerance = default_relative_root_tolerance(ztypes.float)
    function_tolerance = 0

    kernel_instance = kernel(loc=0, scale=bandwidth)

    def objective_fn(x):
        return kernel_instance.prob(x) - tf.constant(absolute_tolerance, ztypes.float)

    roots, value_at_roots, num_iterations, converged = root_search.brentq(
        objective_fn,
        tf.constant(0.0, dtype=ztypes.float),
        tf.constant(8.0, dtype=ztypes.float) * bandwidth,
        absolute_root_tolerance=absolute_root_tolerance,
        relative_root_tolerance=relative_root_tolerance,
        function_tolerance=function_tolerance
    )

    return roots + absolute_root_tolerance


def convolve_1d_data_with_kernel(kernel, bandwidth, data, grid, support = None, fft_method = 'conv1d'):

        kernel_grid_min = tf.math.reduce_min(grid)
        kernel_grid_max = tf.math.reduce_max(grid)

        num_grid_points = tf.size(grid, ztypes.int)
        num_intervals = num_grid_points - tf.constant(1, ztypes.int)
        space_width = kernel_grid_max - kernel_grid_min
        dx = space_width / tf.cast(num_intervals, ztypes.float)

        L = tf.cast(num_grid_points, ztypes.float)

        if support is not None:
            support_bandwidth = support * bandwidth
        else:
            support_bandwidth = find_practical_support_bandwidth(kernel, bandwidth)

        L = tf.minimum(tf.math.floor(support_bandwidth / dx),  L)

        # Calculate the kernel weights
        zero = tf.constant(0, ztypes.float)
        kernel_grid = tf.linspace(zero, dx * L, tf.cast(L, ztypes.int) + tf.constant(1, ztypes.int))
        kernel_weights = kernel(loc=zero, scale=bandwidth).prob(kernel_grid)
        kernel_weights = tf.concat(values=[tf.reverse(kernel_weights, axis = [0])[:-1], kernel_weights], axis=0)

        c = data
        k = kernel_weights

        if fft_method  == 'conv1d':
            c_size = tf.size(c, ztypes.int)
            c = tf.reshape(c, [1, c_size, 1], name='c')

            k_size = tf.size(k, ztypes.int)
            k = tf.reshape(k, [k_size, 1, 1], name='k')

            return tf.squeeze(tf.nn.conv1d(c, k, 1, 'SAME'))

        else:

            P = tf.math.pow(tf.constant(2, ztypes.int), tf.cast(tf.math.ceil(tf.math.log(tf.constant(3.0, ztypes.float) * tf.cast(num_grid_points, ztypes.float) - tf.constant(1.0, ztypes.float)) / tf.math.log(tf.constant(2.0, ztypes.float))), ztypes.int))

            right_padding = tf.cast(P, ztypes.int) - tf.constant(2, ztypes.int) * num_grid_points - tf.constant(1, ztypes.int)
            left_padding = num_grid_points - tf.constant(1, ztypes.int)

            c = tf.pad(data, [[left_padding, right_padding]])
            k = tf.pad(kernel_weights, [[0, right_padding]])

            result = tf.signal.irfft(tf.signal.rfft(c) * tf.signal.rfft(k))
            start, end = tf.constant(2, ztypes.int) * num_grid_points - tf.constant(1, ztypes.int), tf.constant(3, ztypes.int) * num_grid_points - tf.constant(1, ztypes.int)
            result = result[start:end]

            return result



import numpy as np
import pytest


@pytest.mark.parametrize('weights', [None, np.array([1., 2., 3., 4., 5., 6., 7., 8.])])
def test_padreflect_data_weights_1dim(weights):
    from zfit.models.kde import padreflect_data_weights_1dim
    testarray = np.array([0.1, 0.5, -2, 5, 6.7, -1.2, 1.2, 7])

    data, w = padreflect_data_weights_1dim(testarray, mode=0.1, weights=weights)
    true_data = np.array([-2.8, -2, -2, -1.2, 0.1, 0.5, 1.2, 5, 6.7, 7, 7, 7.3])
    np.testing.assert_allclose(true_data, np.sort(data))
    if weights is not None:
        true_w = np.sort(np.array([6, 3, 1., 2., 3., 4., 5., 6., 7., 8., 8, 5]))
        np.testing.assert_allclose(true_w, np.sort(w))

    data, w = padreflect_data_weights_1dim(testarray, mode={'lowermirror': 0.1, 'uppermirror': 0.1}, weights=weights)
    true_data = np.array([-2.8, -2, -2, -1.2, 0.1, 0.5, 1.2, 5, 6.7, 7, 7, 7.3])
    np.testing.assert_allclose(true_data, np.sort(data))
    if weights is not None:
        true_w = np.sort(np.array([6, 3, 1., 2., 3., 4., 5., 6., 7., 8., 8, 5]))
        np.testing.assert_allclose(true_w, np.sort(w))

    data, w = padreflect_data_weights_1dim(testarray, mode={'lowermirror': 0.1, 'uppermirror': 0.3}, weights=weights)
    true_data = np.sort(np.array([-2.8, -2, -2, -1.2, 0.1, 0.5, 1.2, 5, 6.7, 7, 7, 7.3, 9]))
    np.testing.assert_allclose(true_data, np.sort(data))
    if weights is not None:
        true_w = np.sort(np.array([6, 3, 1., 2., 3., 4., 5., 6., 7., 8., 8, 5, 4]))
        np.testing.assert_allclose(true_w, np.sort(w))

    data, w = padreflect_data_weights_1dim(testarray, mode={'lowermirror': 0.1}, weights=weights)
    true_data = np.sort(np.array([-2.8, -2, -2, -1.2, 0.1, 0.5, 1.2, 5, 6.7, 7]))
    np.testing.assert_allclose(true_data, np.sort(data))
    if weights is not None:
        true_w = np.sort(np.array([6, 3, 1., 2., 3., 4., 5., 6., 7., 8., ]))
        np.testing.assert_allclose(true_w, np.sort(w))

    data, w = padreflect_data_weights_1dim(testarray, mode={'lowermirror': 0.3}, weights=weights)
    true_data = np.sort(np.array([-4.1, -4.5, -2.8, -2, -2, -1.2, 0.1, 0.5, 1.2, 5, 6.7, 7]))
    np.testing.assert_allclose(true_data, np.sort(data))
    if weights is not None:
        true_w = np.sort(np.array([1, 2, 6, 3, 1., 2., 3., 4., 5., 6., 7., 8., ]))
        np.testing.assert_allclose(true_w, np.sort(w))

    data, w = padreflect_data_weights_1dim(testarray, mode={'uppermirror': 0.3}, weights=weights)
    true_data = np.sort(np.array([-2, -1.2, 0.1, 0.5, 1.2, 5, 6.7, 7, 7, 7.3, 9]))
    np.testing.assert_allclose(true_data, np.sort(data))
    if weights is not None:
        true_w = np.sort(np.array([1., 2., 3., 4., 5., 6., 7., 8., 8, 5, 4]))
        np.testing.assert_allclose(true_w, np.sort(w))

    with pytest.raises(ValueError):
        data, w = padreflect_data_weights_1dim(testarray, mode={'uper': 0.3}, weights=weights)
    with pytest.raises(ValueError):
        data, w = padreflect_data_weights_1dim(testarray, mode={'uppermirror': 0.3, 'lowe': 0.321}, weights=weights)



"""Numpy like interface for math functions and arrays. This module is intended to replace tensorflow specific methods
and datastructures with equivalent or similar versions in the numpy api. This should help make zfit as a project
portable to alternatives of tensorflow should it be necessary in the future. At the moment it is simply an alias for the
numpy api of tensorflow. See https://www.tensorflow.org/guide/tf_numpy for more a guide to numpy api in tensorflow. See
https://www.tensorflow.org/api_docs/python/tf/experimental/numpy for the complete numpy api in tensorflow. Recommended
way of importing:

>>> Import zfit.z.numpy as znp
"""
from tensorflow.experimental.numpy import *



import numpy as np
import tensorflow as tf
from tf_quant_finance.math import root_search

from .. import z
from ..settings import ztypes
from . import binning as binning_util


@z.function(wraps='tensor')
def calc_f(s, f, squared_integers, grid_data_dct2, N):
    # Step one: estimate t_s from |f^(s+1)|^2
    one_half = tf.constant(1.0 / 2.0, ztypes.float)
    one = tf.constant(1.0, ztypes.float)
    two = tf.constant(2.0, ztypes.float)
    three = tf.constant(3.0, ztypes.float)
    pi = tf.constant(np.pi, ztypes.float)

    odd_numbers_prod = tf.math.reduce_prod(tf.range(one, two * s + one, 2))
    K0 = odd_numbers_prod / tf.math.sqrt(two * pi)
    const = (one + tf.math.pow(one_half,
                               s + one_half)) / three
    time = (tf.math.pow(two
                        * const * K0 / (N * f), two
                        / (three + two * s)))

    # Step two: estimate |f^s| from t_s
    f = (one_half
         * tf.math.pow(pi, (two * s))
         * tf.math.reduce_sum(tf.math.pow(squared_integers, s)
                              * grid_data_dct2 * tf.math.exp(-squared_integers
                                                             * tf.math.pow(pi,
                                                                           two) * time)))

    return f


@z.function(wraps='tensor', autograph=True)
def _fixed_point(t, N, squared_integers, grid_data_dct2):
    r"""
    Compute the fixed point as described in the paper by Botev et al.

    .. math:
        t = \xi \gamma^{5}(t)

    Parameters
    ----------
    t : float
        Initial guess.
    N : int
        Number of data points.
    squared_integers : array-like
        The numbers [1, 2, 9, 16, ...]
    grid_data_dct2 : array-like
        The DCT of the original data, divided by 2 and squared.
    Examples
    --------
    >>> # From the matlab code
    >>> ans = _fixed_point(0.01, 50, np.arange(1, 51), np.arange(1, 51))
    >>> assert np.allclose(ans, 0.0099076220293967618515)
    >>> # another
    >>> ans = _fixed_point(0.07, 25, np.arange(1, 11), np.arange(1, 11))
    >>> assert np.allclose(ans, 0.068342291525717486795)
    References
    ----------
     - Implementation by Daniel B. Smith, PhD, found at
       https://github.com/Daniel-B-Smith/KDE-for-SciPy/blob/master/kde.py
    """

    # ell = 7 corresponds to the 5 steps recommended in the paper
    ell = tf.constant(7, ztypes.float)

    # Fast evaluation of |f^l|^2 using the DCT, see Plancherel theorem
    f = (tf.constant(0.5, ztypes.float)
         * tf.math.pow(tf.constant(np.pi, ztypes.float), (tf.constant(2.0, ztypes.float) * ell))
         * tf.math.reduce_sum(tf.math.pow(squared_integers, ell)
                              * grid_data_dct2
                              * tf.math.exp(-squared_integers
                                            * tf.math.pow(tf.constant(np.pi, ztypes.float),
                                                          tf.constant(2.0, ztypes.float)) * t)))

    i = tf.constant(6., dtype=ztypes.float)
    while_condition = lambda i, f: i > 1

    def body(i, f):
        # do something here which you want to do in your loop
        # increment i
        f = calc_f(i, f, squared_integers, grid_data_dct2, N)
        return i - 1., f

    # do the loop:
    fnew = tf.while_loop(while_condition, body, (i, f), maximum_iterations=5, parallel_iterations=5)[1]

    # This is the minimizer of the AMISE
    t_opt = tf.math.pow(tf.constant(2 * np.sqrt(np.pi), ztypes.float)
                        * N * fnew, tf.constant(-2.0 / 5.0, ztypes.float))

    # Return the difference between the original t and the optimal value
    return t - t_opt


def _find_root(function, N, squared_integers, grid_data_dct2):
    """Root finding algorithm. Based on MATLAB implementation by Botev et al.

    >>> # From the matlab code
    >>> ints = np.arange(1, 51)
    >>> ans = _root(_fixed_point, N=50, args=(50, ints, ints))
    >>> np.allclose(ans, 9.237610787616029e-05)
    True
    """
    # From the implementation by Botev, the original paper author
    # Rule of thumb of obtaining a feasible solution
    N2 = tf.math.maximum(tf.math.minimum(tf.constant(1050, ztypes.float), N), tf.constant(50, ztypes.float))
    tol = 10e-12 + 0.01 * (N2 - 50) / 1000
    left_bracket = tf.constant(0.0, dtype=ztypes.float)
    right_bracket = tf.constant(10e-12, ztypes.float) + tf.constant(0.01, ztypes.float) * \
                    (N2 - tf.constant(50, ztypes.float)) / tf.constant(1000, ztypes.float)

    converged = tf.constant(False)
    t_star = tf.constant(0.0, dtype=ztypes.float)

    def fixed_point_function(t):
        return _fixed_point(t, N, squared_integers, grid_data_dct2)

    def condition(right_bracket, converged, t_star): return tf.math.logical_not(converged)

    def body(right_bracket, converged, t_star):
        t_star, value_at_t_star, num_iterations, converged = root_search.brentq(
            fixed_point_function,
            left_bracket,
            right_bracket,
            None,
            None,
            2e-12
        )

        t_star = t_star - value_at_t_star

        right_bracket = right_bracket * tf.constant(2.0, ztypes.float)

        return right_bracket, converged, t_star

    # While a solution is not found, increase the tolerance and try again
    right_bracket, converged, t_star = tf.while_loop(condition, body, [right_bracket, converged, t_star])

    return t_star


def _calculate_t_star(data, num_grid_points, binning_method, weights):
    # Setting `percentile` higher decreases the chance of overflow
    grid = binning_util.generate_1d_grid(data, num_grid_points, 6.0, 0.5)

    # Create an equidistant grid
    R = tf.cast(tf.reduce_max(data) - tf.reduce_min(data), ztypes.float)

    # dx = R / tf.constant((num_grid_points - 1), ztypes.float)
    data_unique, data_unique_indexes = tf.unique(data)
    N = tf.cast(tf.size(data_unique), ztypes.float)

    # Use linear binning to bin the data on an equidistant grid, this is a
    # prerequisite for using the FFT (evenly spaced samples)
    grid_data = binning_util.bin_1d(binning_method, data, grid, weights)

    # Compute the type 2 Discrete Cosine Transform (DCT) of the data
    grid_data_dct = tf.signal.dct(grid_data, type=2)

    # Compute the bandwidth
    squared_integers = tf.math.pow(
        tf.range(1, num_grid_points, dtype=ztypes.float), tf.constant(2, ztypes.float))
    grid_data_dct2 = tf.math.pow(grid_data_dct[1:], 2) / 4

    # Solve for the optimal (in the AMISE sense) t
    t_star = _find_root(_fixed_point, N, squared_integers, grid_data_dct2)

    return t_star, R, squared_integers, grid_data_dct, grid


def _calculate_density(t_star, R, squared_integers, grid_data_dct):
    # Prepend zero
    squared_integers = tf.pad(squared_integers, [[1, 0]])

    # Smooth the initial data using the computed optimal t
    # Multiplication in frequency domain is convolution
    grid_data_dct_t = grid_data_dct * tf.math.exp(-squared_integers * tf.math.pow(tf.constant(
        np.pi, ztypes.float), tf.constant(2.0, ztypes.float)) * t_star / tf.constant(2.0, ztypes.float))

    # Diving by 2 done because of the implementation of tf.signal.idct
    density = tf.signal.idct(grid_data_dct_t, type=2) / (2 * R)

    # Due to overflow, some values might be smaller than zero, correct it
    density = tf.cast(density > 0, density.dtype) * density

    return density


def calculate_bandwidth(data, num_grid_points=1024, binning_method='linear', weights=None):
    data = tf.cast(data, ztypes.float)

    t_star, R, squared_integers, grid_data_dct, grid = _calculate_t_star(
        data, num_grid_points, binning_method, weights)

    return tf.math.sqrt(t_star) * R


def calculate_bandwidth_and_density(data, num_grid_points=1024, binning_method='linear', weights=None):
    data = tf.cast(data, ztypes.float)

    t_star, R, squared_integers, grid_data_dct, grid = _calculate_t_star(
        data, num_grid_points, binning_method, weights)

    bandwidth = tf.math.sqrt(t_star) * R
    density = _calculate_density(t_star, R, squared_integers, grid_data_dct)

    return bandwidth, density, grid
